
/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/

/*
 * This is a SAMPLE program for illustration purposes only.
 * IBM makes no commitment for support of this module.
 */

/*
===============================================================================
	Common Graphics	Environment (CGE) extensions 
	PEX 5.1	Simple Programming Example

	Main program file

===============================================================================
*/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/extensions/multibuf.h>

#include <X11/PEX5/PEXlib.h>
#include <X11/PEX5/PEXExtlib.h>

#include <PEXUtCmap.h>

#include "ptorus.h"


/*
*******************************************************************************
    Forward declarations of utility procedures in this file.
*******************************************************************************
*/

static void redraw(
#if NeedFunctionPrototypes
	Display	*dpy,
	Multibuffer *drawable,
	PEXRenderer renderer,
	PEXStructure structure,
	Multibuffer *buffers
#endif
);

static void p_animate(
#if NeedFunctionPrototypes
	Display	*dpy,
	PEXRenderer rid,
	PEXStructure sid
#endif
);


/*
*******************************************************************************
    This data structure	describes the enumerated type values this program
    requires.  They are	checked	in the main program, below.
    Note that the only check we	really need to make is to look for
    the	CGE Escape Identifiers.	 If they are present, we can be	sure that
    we have a CGE server and can assume	that a large number of features
    are	supported.
    This program requires no features beyond what CGE PEX 5.1 requires,
    so we just need to check for CGE support.
*******************************************************************************
*/

static struct enum_entry {
    unsigned int	enum_type;
    unsigned int	enum_value;
    char		*message;
} enums_to_check[]={

    { PEXETEscape,		PEXExtEscapeChangePipelineContext,
	"CGE PEX 5.1 Support" },

};
#define	NUM_ENUMS_TO_CHECK (sizeof(enums_to_check)/sizeof(struct enum_entry))


/*
*******************************************************************************
    The	main program.  Takes one possible command-line argument	to name
    the	X server, -display <display_name>.
*******************************************************************************
*/

main (argc, argv)
	int argc;
	char **argv;
{
    char			*prog_name;
    char			*display_name;

    int				screen;

    Display			*dpy;
    Window			window;
    Multibuffer			renderBuf;
    Multibuffer			buffers[2];
    int				mbx_buf_count;

    PEXExtensionInfo		*ext_info;
    char			err_string[PEXErrorStringLength];

    XVisualInfo			vis_info;
    XStandardColormap		cmap_info;
    PEXColorApproxEntry		capx_info;

    XSizeHints			hints;
    PEXUtVisualCriteria		criteria;
    PEXUtWindowSpecification	window_info;
    int				crit_index;
    unsigned int		unmet;
    Atom			prop_atom;
    int				result;
    XColor			returned_background;
    XEvent			event;
    Colormap			cmap_id;

    Atom			xa_WM_DELETE_WINDOW;
    Atom			xa_MOTIF_WM_MESSAGES;
    Atom			xa_WM_PROTOCOLS;
    Atom			protocols[2];
    char			*title;

    PEXRendererAttributes	rend_attrs;
    PEXRenderer			renderer;
    PEXStructure		structure;

    int				num_lights;
    PEXLightEntry		*lights;
    PEXTableIndex		*lss;

    int				visible;
    camera_arg			camera;


    /*
	Get the	program	name for error messages.
    */
    if (prog_name = strrchr(argv[0], '/'))
	prog_name++;
    else
	prog_name = argv[0];


    /*
	Parse the command line arguments.
    */
    if ((argc >	2)
	    && argv[1]
	    && !strcmp(	"-display", argv[1] ))
	display_name = XDisplayName( argv[2] );
    else	
	display_name = XDisplayName( NULL );


    /*
	Open the X connection.
    */
    if ( !(dpy = XOpenDisplay( display_name )) )
    {
	fprintf(stderr,	"%s: can't open	display	%s -- exiting\n",
			prog_name, display_name);
	exit (1);
    }

    screen = DefaultScreen(dpy);

    /*
	Initialize PEXlib on the connection.
    */
    if (PEXInitialize (dpy, &ext_info, PEXErrorStringLength, err_string))
    {
	fprintf	(stderr, "%s: PEXInitialize() failed on	%s: %s\n",
		 prog_name, display_name, err_string);
	exit (1);
    }


    /*
	Select a Visual	for PEX	rendering and create a Window
	on the Visual.
	We're going to hint that we want Double	Buffering for
	smooth animation and to	use a Standard Colormap	to
	promote	the sharing of Colormap	resources on the server.
	If we don't get	either or both,	we'll stumble on.
    */

    hints.x = 128;
    hints.y = 128;
    hints.width	= 256;
    hints.height = 256;
    hints.flags	= (USSize|USPosition);

    criteria.hard_criteria_mask	= 0;
    criteria.soft_criteria_mask	= PEXUtDBCapability |
				  PEXUtStandardColormapProperty;

    criteria.double_buffering_capability = PEXUtDBPEX;
    criteria.standard_colormap_property	= True;

    window_info.attr_mask = CWEventMask;
    window_info.attrs.event_mask = ( ExposureMask	 |
				    VisibilityChangeMask |
				    StructureNotifyMask);

    window_info.title =	"PEX Torus";
    window_info.size_hints = hints;
    window_info.parent = RootWindow (dpy, screen);
    window_info.border_width = 0;
    window_info.background_color_name =	"black";
    window_info.border_color_name = "white";

    result = PEXUtMakeWindowAndColormap	(dpy, screen,
					   1, &criteria,
					   &window_info, &window,
					   &vis_info, &cmap_info, &capx_info,
					   &crit_index,	&unmet,	&prop_atom,
					   &returned_background,
					   &cmap_id);

    if (result != PEXUtSuccess)
    {
	static char *PEXUtFailStrings[]	=
	{
	    "PEXUtBadAlloc",		    /* -4 maps to 0 */
	    "PEXUtPEXFailure",		    /* -3 maps to 1 */
	    "PEXUtXFailure",		    /* -2 maps to 2 */
	    "PEXUtCriteriaFailure",	    /* -1 maps to 3 */
	    "PEXUtSuccess",		    /*	0 maps to 4 */
	    "PEXUtQualifiedSuccess",	    /* +1 maps to 5 */
	    "PEXUtAlternativeSuccess",	    /* +2 maps to 6 */
	};

	fprintf(stderr,"%s: visual selection or	window creation	failed:	%s\n",
		prog_name, PEXUtFailStrings[ result + 4	]);
	exit(1);
    }

    /*
       Report soft criteria failures, but don't	quit.
    */
    if (unmet &	PEXUtDBCapability)
	fprintf(stderr,
		"%s: could not find Visual for double buffering\n",
		 prog_name);
    if (unmet &	PEXUtStandardColormapProperty)
	fprintf(stderr,
		"%s: could not find Standard Colormap Property\n",
		prog_name);

    /*
	Set up some window manager properties (see the ICCCM).
    */

    xa_WM_DELETE_WINDOW	 = XInternAtom(dpy, "WM_DELETE_WINDOW"	  , False);
    xa_MOTIF_WM_MESSAGES = XInternAtom(dpy, "_MOTIF_WM_MESSAGES"  , False);
    xa_WM_PROTOCOLS	 = XInternAtom(dpy, "WM_PROTOCOLS"	  , False);
    protocols[0] = xa_WM_DELETE_WINDOW ;
    protocols[1] = xa_MOTIF_WM_MESSAGES;
    XSetWMProtocols(dpy, window, protocols, 2);

    title = "PEX Torus";
    XChangeProperty(dpy, window, XA_WM_ICON_NAME, XA_STRING,
		    8, PropModeReplace,	(unsigned char *) title, 8);


    /*
	Configure the window for double-buffering via MBX.
    */

    mbx_buf_count = XmbufCreateBuffers(	dpy, window, 2,
			      MultibufferUpdateActionBackground,
			      MultibufferUpdateHintFrequent,
			      buffers );
    if ( mbx_buf_count == 0 ) {
	fprintf(stderr,
	    "%s: can't create MBX buffers on display %s, visual	%x\n",
		prog_name, display_name, vis_info.visualid);
	exit(1);
    }

    /* 
	Handle case where only one MBX buffer was created.
	This will make the swap	work, even if one buffer.
    */
    if ( mbx_buf_count == 1 )
      buffers[1] = buffers[0];

    renderBuf =	buffers[1];


    /*
	Check the Enum Types to	be sure	that CGE PEX is	supported
    */
    {
	int status, i, j, found;
	int enum_types[NUM_ENUMS_TO_CHECK];
	unsigned long *enum_counts;
	PEXEnumTypeDesc	*enum_values, *p_enum;
	
	for (i=0; i<NUM_ENUMS_TO_CHECK;	i++)
	    enum_types[i] = enums_to_check[i].enum_type;

	status = PEXGetEnumTypeInfo (dpy, renderBuf,
			    NUM_ENUMS_TO_CHECK,	enum_types,
			    PEXETCounts|PEXETIndex,
			    &enum_counts, &enum_values);
	if (!status) {
	    fprintf(stderr,
		"%s: inquiry of	enumerated types failed	- exiting\n",
		prog_name);
	    exit(1);
	}

	p_enum = enum_values;
	for (i=0; i<NUM_ENUMS_TO_CHECK;	i++) {

	    found = False;

	    for	(j=0; j<enum_counts[i];	j++) {
		if (((unsigned short) (p_enum++)->index) ==
		    ((unsigned short) enums_to_check[i].enum_value)) {
		    found = True;
		}
	    }

	    if (!found)	{
		fprintf(stderr,
		    "%s: this program requires support for %s\n",
		    prog_name, enums_to_check[i].message);
		exit(1);
	    }
	}

	PEXFreeEnumInfo	(NUM_ENUMS_TO_CHECK, enum_counts, enum_values);
    }


    /*
	Create the model and scene to be animated.
    */

    structure =	create_model(dpy);
    create_scene(dpy, &camera, &num_lights, &lights, &lss);


    /*
	Create a Renderer and various lookup tables.
	Set up the view, lights, and Pipeline Context.
    */

    renderer = create_renderer(dpy, window, &rend_attrs);

    set_view_camera(dpy, window, &camera, renderer, &rend_attrs, 1);
    PEXSetTableEntries(dpy, rend_attrs.light_table, lss[0], num_lights,
			PEXLUTLight, lights);
    PEXSetTableEntries(dpy, rend_attrs.color_approx_table, 0, 1,
			PEXLUTColorApprox, &capx_info);
    create_initial_state (dpy, rend_attrs.pipeline_context, num_lights,	lss);



    /*
	Now enter the main event-processing and	animation loop.
    */

    visible = False;
    while (True)
    {

	/*
	    First handle any pending events.
	*/

	while (XPending(dpy))
	{
	    XNextEvent(	dpy, &event );

	    switch ( event.type	) {
	    case ConfigureNotify:
		/*
		    Adapt the view to the window configuration
		    and	render the image.
		*/
		set_view_camera(dpy, window, &camera, renderer,	&rend_attrs, 1);
		redraw(dpy, &renderBuf,	renderer, structure, buffers);
		break;

	    case Expose:
		/*
		    Flush remaining Expose events.
		*/
		while (	XCheckTypedWindowEvent(	dpy, window, Expose, &event ))
		    ;		/* empty statement */
		visible	= True;

		/*
		    Adapt the view to the window configuration
		    and	render the image.
		*/
		set_view_camera(dpy, window, &camera, renderer,	&rend_attrs, 1);
		redraw(dpy, &renderBuf,	renderer, structure, buffers);
		break;

	    case ClientMessage:
		if (event.xclient.message_type == xa_WM_PROTOCOLS)
		{
		    if (event.xclient.data.l[0]	== xa_WM_DELETE_WINDOW)
		    {
			/*
			    Terminate the double-buffering on the window.
			    Close the Window and the connection.
			*/

			XmbufDestroyBuffers( dpy, window );
			XDestroyWindow (dpy, window);
			XCloseDisplay( dpy );
			exit(0);
		    }
		}
		break;
	    }
	} /* event handling */


	/*
	    After handling events, render the next frame in the	animation.
	*/

	if (visible)
	    redraw(dpy,	&renderBuf, renderer, structure, buffers);

    } /* main loop */

} /* main */


/*
*******************************************************************************
    This procedure draws one frame and swaps the buffers.
*******************************************************************************
*/
static void redraw(dpy,	drawable, renderer, structure, buffers)
	Display	*dpy;
	Multibuffer *drawable;
	PEXRenderer renderer;
	PEXStructure structure;
	Multibuffer *buffers;
{
    PEXBeginRendering(dpy, (Drawable) *drawable, renderer);
    p_animate(dpy, renderer, structure);
    PEXEndRendering(dpy, renderer, True);

    /* Swap the	buffers. */
    XmbufDisplayBuffers( dpy, 1, drawable, 0, 0	);
    *drawable =	((*drawable == buffers[0]) ? buffers[1]	: buffers[0]);

    XSync(dpy, False);

} /* redraw */


/*
*******************************************************************************
    This procedure generates the animation and does the	rendering
    of a single	frame.
*******************************************************************************
*/

static float xangle = 0.01;
static float yangle = 0.1;

static void p_animate(dpy, rid,	sid)
	Display	*dpy;
	PEXRenderer rid;
	PEXStructure sid;
{
    PEXMatrix model_mtx, xrot_mtx, yrot_mtx;

    xangle += 0.01;
    if (xangle > 2*M_PI) xangle	= 0.01;
    PEXRotate(PEXXAxis,	xangle,	xrot_mtx);

    yangle += 0.1;
    if (yangle > 2*M_PI) yangle	= 0.1;
    PEXRotate(PEXYAxis,	yangle,	yrot_mtx);

    PEXMatrixMult (xrot_mtx, yrot_mtx, model_mtx);

    PEXSetLocalTransform(dpy, rid, PEXOCRender,	PEXReplace, model_mtx);
    PEXExecuteStructure(dpy, rid, PEXOCRender, sid);

} /* p_animate */
